int find_channel_by_name(unsigned server, const char *channel)
{
  unsigned n = servers[server].channel_count;

  if(!channel)
    return -1;

  printf("Active channel %s\n", channel);

  while(n--)
    if(!strcmp(channel, servers[server].channels[n].name))
      break;

  return n;
}

void channel_set_topic(unsigned server, char *cmdptr)
{
  char channel[CHANLEN];
  int n;

  while(*cmdptr != ' ') cmdptr++; // nick
  cmdptr++;

  for(n=0;*cmdptr != ' ';n++)
    channel[n] = *cmdptr++;
  channel[n] = 0;

  n = find_channel_by_name(server, channel);
  if(n<0) return;

  if(*++cmdptr == ':') cmdptr++;

  cmdptr[strnllen(cmdptr)] = 0;
  strcpy(servers[server].channels[n].topic, cmdptr);
}

void channel_reset_topic(unsigned server, char *cmdptr)
{
  char channel[CHANLEN];
  char nick[NICKLEN];
  int n;
  char msg[100];

  for(n=0;*cmdptr != ' ';n++)
    nick[n] = *cmdptr++;
  nick[n] = 0;

  cmdptr += 7;

  for(n=0;*cmdptr != ' ';n++)
    channel[n] = *cmdptr++;
  channel[n] = 0;

  n = find_channel_by_name(server, channel);
  if(n<0) return;

  if(*++cmdptr == ':') cmdptr++;

  cmdptr[strnllen(cmdptr)] = 0;
  strcpy(servers[server].channels[n].topic, cmdptr);

  sprintf(msg, "%s updated channel topic.\n", nick);
  push_channel_message(server, msg, strlen(msg), n);

  if(active_server == server && servers[server].active_channel == n)
    show_topic(servers[server].channels[n].topic);
}

void channel_add_nick(unsigned server, unsigned channel, const char *nick)
{
  void *ptr;

  pthread_mutex_lock(&lock);

  if(!servers[server].channels[channel].user_count)
  {
    servers[server].channels[channel].users = malloc(++servers[server].channels[channel].user_count * sizeof(char*));
    if(!servers[server].channels[channel].users)
    {
      nomemory:
      servers[server].channels[channel].user_count--;
      pthread_mutex_unlock(&lock);
      return;
    }
  }
  else
  {
    ptr = realloc(servers[server].channels[channel].users, ++servers[server].channels[channel].user_count*sizeof(char*));
    if(!ptr) goto nomemory;
    servers[server].channels[channel].users = ptr;
  }

  printf("Adding name %s\n", nick);
  servers[server].channels[channel].users[servers[server].channels[channel].user_count-1] = malloc((strlen(nick)+1) * sizeof(char));
  strcpy(servers[server].channels[channel].users[servers[server].channels[channel].user_count-1], nick);

  pthread_mutex_unlock(&lock);
}

void channel_del_nick(unsigned server, unsigned channel, const char *nick)
{
  int n = servers[server].channels[channel].user_count;
  char msg[100];

  while(n--)
    if(servers[server].channels[channel].users[n][0] == '@' || servers[server].channels[channel].users[n][0] == '+')
    {
      if(!strcmp(&servers[server].channels[channel].users[n][1], nick))
        break;
    }
    else if(!strcmp(servers[server].channels[channel].users[n], nick))
      break;

  if(n < 0) return;

  pthread_mutex_lock(&lock);

  free(servers[server].channels[channel].users[n]);
  servers[server].channels[channel].user_count--;

  while(n < servers[server].channels[channel].user_count)
    servers[server].channels[channel].users[n] = servers[server].channels[channel].users[n+1];

  servers[server].channels[channel].users = realloc(servers[server].channels[channel].users, servers[server].channels[channel].user_count * sizeof(char*));

  pthread_mutex_unlock(&lock);

  sprintf(msg, "%s has left the channel.\n", nick);
  push_channel_message(server, msg, strlen(msg), channel);
}

void add_user_to_channel(unsigned server, char *nickptr, char *chanptr)
{
  char nick[NICKLEN];
  char chan[CHANLEN];
  unsigned n;
  char msg[100];
  unsigned channel;

  for(n=0;nickptr[n] != '!';n++)
    nick[n] = nickptr[n];
  nick[n] = 0;

  for(n=0;chanptr[n] > ' ';n++)
    chan[n] = chanptr[n];
  chan[n] = 0;

  channel_add_nick(server, channel = find_channel_by_name(server, chan), nick);
  update_ui();

  sprintf(msg, "%s has joined the channel.\n", nick);
  push_channel_message(server, msg, strlen(msg), channel);
}

void remove_user_from_channel(unsigned server, char *nickptr, char *chanptr)
{
  char nick[NICKLEN];
  char chan[CHANLEN];
  unsigned n;
  int chanid;

  for(n=0;nickptr[n] != '!';n++)
    nick[n] = nickptr[n];
  nick[n] = 0;

  for(n=0;chanptr[n] > ' ';n++)
    chan[n] = chanptr[n];
  chan[n] = 0;

  chanid = find_channel_by_name(server, chan);
  if(chanid < 0) return;

  channel_del_nick(server, chanid, nick);
  update_ui();
}
