#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>

#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <pthread.h>
#include <unistd.h>
#include <ctype.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <SDL2/SDL.h>
#include <opus/opus.h>
#include <poll.h>

#include "snddev.c"
#include "defines.c"
#include "structs.c"
#include "globals.c"
#include "settings.c"
#include "channels.c"
#include "useracts.c"
#include "voip.c"
#include "messages.c"
#include "sockets.c"
#include "servers.c"
#include "setwin.c"
#include "ui.c"

void append_user_message(const char *text)
{
  char new_text[512];
  const char *user_message = gtk_entry_get_text((GtkEntry*)my_message);

  sprintf(new_text, "%s%s", user_message, text);
  gtk_entry_set_text((GtkEntry*)my_message, new_text);

  gtk_entry_grab_focus_without_selecting((GtkEntry*)my_message);
  gtk_editable_set_position((GtkEditable*)my_message, -1);

}

void set_user_message(const char *text)
{
  gtk_entry_set_text((GtkEntry*)my_message, text);

  gtk_entry_grab_focus_without_selecting((GtkEntry*)my_message);
  gtk_editable_set_position((GtkEditable*)my_message, -1);
}

void empty_channel_list(void)
{
  void *ptr;
  unsigned n;

  while(ptr = gtk_list_box_get_row_at_index((GtkListBox*)channel_list, 0))
    gtk_container_remove((GtkContainer*)channel_list, (GtkWidget*)ptr);

  if(active_server >= 0)
  {
    servers[active_server].label = 0;
    for(n=servers[active_server].channel_count;n--;)
      servers[active_server].channels[n].label = 0;
    for(n=servers[active_server].query_count;n--;)
      servers[active_server].queries[n].label = 0;
  }
}


void update_channel_list(void)
{
  unsigned n;
  GtkTextIter iter;

  if(active_server < 0)
  {
    empty_channel_list();
    list_settings();
    return;
  }

  if(!servers[active_server].label)
  {
    servers[active_server].label = gtk_label_new(servers[active_server].server_name);
    gtk_label_set_attributes(GTK_LABEL(servers[active_server].label), ms_attrlist);
    gtk_list_box_insert((GtkListBox*)channel_list, servers[active_server].label, 0);
    if(!servers[active_server].has_tab)
    {
      gtk_notebook_append_page((GtkNotebook*)tabs, gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0), gtk_label_new(servers[active_server].server_name));
      servers[active_server].has_tab = 1;
    }
    if(!servers[active_server].messages.messages)
    {
      servers[active_server].messages.messages = gtk_text_buffer_new(0);
      gtk_text_buffer_create_tag(servers[active_server].messages.messages, "monospace", "family", "Mono", 0);
      gtk_text_buffer_get_start_iter(servers[active_server].messages.messages, &iter);
      servers[active_server].messages.mark = gtk_text_buffer_create_mark(servers[active_server].messages.messages, 0, &iter, 0);
    }
  }
  for(n=servers[active_server].channel_count;n--;)
  {
    if(!servers[active_server].channels[n].label)
    {
      servers[active_server].channels[n].label = gtk_label_new(servers[active_server].channels[n].name);
      gtk_label_set_xalign((GtkLabel*)servers[active_server].channels[n].label, 0);
      gtk_label_set_attributes(GTK_LABEL(servers[active_server].channels[n].label), ms_attrlist);
//	      gtk_list_box_insert((GtkListBox*)channel_list, servers[active_server].channels[n].label, -1);
      gtk_list_box_insert((GtkListBox*)channel_list, servers[active_server].channels[n].label, servers[active_server].channel_count);
      if(!servers[active_server].channels[n].messages.messages)
      {
        servers[active_server].channels[n].messages.messages = gtk_text_buffer_new(0);
        gtk_text_buffer_create_tag(servers[active_server].channels[n].messages.messages, "monospace", "family", "Mono", 0);
        gtk_text_buffer_get_start_iter(servers[active_server].channels[n].messages.messages, &iter);
        servers[active_server].channels[n].messages.mark = gtk_text_buffer_create_mark(servers[active_server].channels[n].messages.messages, 0, &iter, 0);
      }
    }
  }
  for(n=servers[active_server].query_count;n--;)
  {
    if(!servers[active_server].queries[n].label)
    {
      servers[active_server].queries[n].label = gtk_label_new(servers[active_server].queries[n].nick);
      gtk_label_set_attributes(GTK_LABEL(servers[active_server].queries[n].label), ms_attrlist);
      gtk_list_box_insert((GtkListBox*)channel_list, servers[active_server].queries[n].label, -1);
    }
  }

  gtk_widget_show_all(channel_list);
  gtk_widget_show_all(tabs);
}

void empty_user_list(void)
{
  void *ptr;
  while(ptr = gtk_list_box_get_row_at_index((GtkListBox*)nick_list, 0))
    gtk_container_remove((GtkContainer*)nick_list, (GtkWidget*)ptr);
}

void update_user_list(void)
{
  unsigned n;// = servers[active_server].channels[servers[active_server].active_channel].user_count;
  void *ptr;

//  puts("Updating user list...");

  if(active_server < 0)
    return;

  empty_user_list();

  if(servers[active_server].active_channel < 0)
  {
    list_server_commands();
  }

  else if(servers[active_server].active_channel < servers[active_server].channel_count)
  {
    n = servers[active_server].channels[servers[active_server].active_channel].user_count;
    while(n--)
    {
      ptr = gtk_label_new(servers[active_server].channels[servers[active_server].active_channel].users[n]);
      gtk_label_set_xalign((GtkLabel*)ptr, 1);
      gtk_label_set_attributes(GTK_LABEL(ptr), ms_attrlist);
      gtk_list_box_insert((GtkListBox*)nick_list, (GtkWidget*)ptr, -1);
    }
  }

  bold_name(speaking_name);
  speaking_name = 0;

  gtk_widget_show_all(nick_list);
}

void scroll_down(void)
{
  if(active_server >= 0)
  {
    if(servers[active_server].active_channel < 0)
      gtk_text_view_scroll_to_mark((GtkTextView*)messages_box, (GtkTextMark*)servers[active_server].messages.mark, 0, 0, 0, 0);
    else if(servers[active_server].active_channel < servers[active_server].channel_count)
      gtk_text_view_scroll_to_mark((GtkTextView*)messages_box, (GtkTextMark*)servers[active_server].channels[servers[active_server].active_channel].messages.mark, 0, 0, 0, 0);
    else
      gtk_text_view_scroll_to_mark((GtkTextView*)messages_box, (GtkTextMark*)servers[active_server].queries[servers[active_server].channel_count-servers[active_server].active_channel].messages.mark, 0, 0, 0, 0);

  }

  else
  {
    gtk_text_view_scroll_to_mark((GtkTextView*)messages_box, console_messages.mark, 0, 0, 0, 0);
  }

  puts("Scrolling down...");
}

//int idle_function(void *arg)
int idle_function(GIOChannel *src, GIOCondition cond, void *data)
{
  char message[MSGLEN];
  int message_len;
  GtkTextIter iter;
  unsigned char signal;
  unsigned index;

  read(pipefd[0], &signal, 1);

  pthread_mutex_lock(&lock);

  switch(signal)
  {
/*    case 'u':
      update_user_list();*/
      break;
    default: goto update_all;
  }

  pthread_mutex_unlock(&lock);
  return 0;

  update_all:

//  puts("Updating UI...");

//  puts("Updating channel list...");
  update_channel_list();
//  puts("Updating user list...");
  update_user_list();

  if(active_server < 0)
  {
    while((message_len = get_channel_message(active_server, message, 0)) >= 0)
    {
      gtk_text_buffer_get_iter_at_mark(console_messages.messages, &iter, console_messages.mark);
      gtk_text_buffer_insert_with_tags_by_name(console_messages.messages, &iter, message, message_len, "monospace", 0);
    }
    select_channel();
    scroll_down();
  }

  else
  {
    if(gtk_list_box_get_row_at_index((GtkListBox*)channel_list, servers[active_server].active_channel+1) != gtk_list_box_get_selected_row((GtkListBox*)channel_list))
      select_channel();
    while((message_len = get_channel_message(active_server, message, servers[active_server].active_channel)) >= 0)
    {
      if(servers[active_server].active_channel < 0)
      {
        puts("Writing to server window...");
        gtk_text_buffer_get_iter_at_mark(servers[active_server].messages.messages, &iter, servers[active_server].messages.mark);
        puts("Text iterator initialized");
        gtk_text_buffer_insert_with_tags_by_name(servers[active_server].messages.messages, &iter, message, message_len, "monospace", 0);
      }
      else if(servers[active_server].active_channel < servers[active_server].channel_count)
      {
        puts("Writing to channel window...");
        gtk_text_buffer_get_iter_at_mark(servers[active_server].channels[servers[active_server].active_channel].messages.messages, &iter, servers[active_server].channels[servers[active_server].active_channel].messages.mark);
        puts("Text iterator initialized");
        gtk_text_buffer_insert_with_tags_by_name(servers[active_server].channels[servers[active_server].active_channel].messages.messages, &iter, message, message_len, "monospace", 0);
      }
      else
      {
        index = servers[active_server].channel_count - servers[active_server].active_channel;
        puts("Writing to channel window...");
        gtk_text_buffer_get_iter_at_mark(servers[active_server].queries[index].messages.messages, &iter, servers[active_server].queries[index].messages.mark);
        puts("Text iterator initialized");
        gtk_text_buffer_insert_with_tags_by_name(servers[active_server].queries[index].messages.messages, &iter, message, message_len, "monospace", 0);
      }
      scroll_down();
    }
  }

//  puts("Updated UI.");

  pthread_mutex_unlock(&lock);

  return 1;
}

int select_channel(void)
{
  int channel;
  int query;
  unsigned n;
  char *selection_commands[] =
  {
    "/scall ",
    "/pcall "
  };
  char *selected_text = (char*)gtk_label_get_text((GtkLabel*)gtk_bin_get_child((GtkBin*)gtk_list_box_get_selected_row((GtkListBox*)channel_list)));

  if(!selected_text)
  {
    if(active_server < 0)
      gtk_list_box_select_row((GtkListBox*)channel_list, gtk_list_box_get_row_at_index((GtkListBox*)channel_list, 0));
    else
      gtk_list_box_select_row((GtkListBox*)channel_list, gtk_list_box_get_row_at_index((GtkListBox*)channel_list, servers[active_server].active_channel+1));
    selected_text = (char*)gtk_label_get_text((GtkLabel*)gtk_bin_get_child((GtkBin*)gtk_list_box_get_selected_row((GtkListBox*)channel_list)));
    if(!selected_text) return 0;
  }

  if(active_server == -1)
  {
    empty_user_list();
    gtk_text_view_set_buffer((GtkTextView*)messages_box, console_messages.messages);
    for(n=sizeof(command_center_tabs)/sizeof(void*[2]);n--;)
      if(!strcmp(command_center_tabs[n][0], selected_text))
      {
        list_options(command_center_tabs[n][1]);
        break;
      }
    return 0;
  }

  for(n=sizeof(selection_commands)/sizeof(char*);n--;)
    if(!strncmp(gtk_entry_get_text((GtkEntry*)my_message), selection_commands[n], sizeof(selection_commands[n])))
    {
      append_user_message(gtk_label_get_text((GtkLabel*)gtk_bin_get_child((GtkBin*)gtk_list_box_get_selected_row((GtkListBox*)channel_list))));
      break;
    }

  channel = find_channel_by_name(active_server, selected_text);
  if(channel < 0)
  {
    query = find_server_query(active_server, selected_text);
    if(query >= 0)
      channel = servers[active_server].channel_count + query;
  }

  servers[active_server].active_channel = channel;
  if(channel >= 0)
  {
    if(channel < servers[active_server].channel_count)
    {
      show_topic(servers[active_server].channels[channel].topic);
      gtk_text_view_set_buffer((GtkTextView*)messages_box, servers[active_server].channels[channel].messages.messages);
    }
    else
    {
      show_topic(servers[active_server].queries[query].nick);
      gtk_text_view_set_buffer((GtkTextView*)messages_box, servers[active_server].queries[query].messages.messages);
    }
  }
  else serverwindow:
    gtk_text_view_set_buffer((GtkTextView*)messages_box, servers[active_server].messages.messages);

  update_ui();
}

int select_server(void)
{
  pthread_mutex_lock(&lock);
  empty_channel_list();
  empty_user_list();
  pthread_mutex_unlock(&lock);
  active_server = server_count-1-gtk_notebook_get_current_page((GtkNotebook*)tabs);
  printf("Active server %i\n", active_server);
  update_ui();
//  select_channel();
}

int select_nick(void)
{
  const char *user_message = gtk_entry_get_text((GtkEntry*)my_message);
  unsigned selection = gtk_list_box_row_get_index(gtk_list_box_get_selected_row((GtkListBox*)nick_list));
  char new_text[512];

  if(active_server == -1)
  {
    select_option(selection);
  }
  else
    if(servers[active_server].active_channel == -1)
    {
      if(strlen(user_message) < 32)
      {
        gtk_entry_set_text((GtkEntry*)my_message, server_commands[selection][1]);
        push_channel_message(active_server, server_commands[selection][2], strlen(server_commands[selection][2]), -1);
      }
    }
    else
    {
      if(user_message[0])
        append_user_message(gtk_label_get_text((GtkLabel*)gtk_bin_get_child((GtkBin*)gtk_list_box_get_selected_row((GtkListBox*)nick_list))));
      else
      {
        sprintf(new_text, "%s: ", gtk_label_get_text((GtkLabel*)gtk_bin_get_child((GtkBin*)gtk_list_box_get_selected_row((GtkListBox*)nick_list))));
        gtk_entry_set_text((GtkEntry*)my_message, new_text);
      }
    }

  gtk_entry_grab_focus_without_selecting((GtkEntry*)my_message);
  gtk_editable_set_position((GtkEditable*)my_message, -1);
}

void activate(GtkApplication *app, GApplicationCommandLine *command_line, gpointer user_data)
{
  int argc;
  GError *error;
  GtkTextIter iter;

  // set monospace font attributes
  ms_attrlist = pango_attr_list_new();
  ms_df = pango_font_description_new();
  pango_font_description_set_family(ms_df, "Mono");
  pango_font_description_set_size(ms_df, 10*PANGO_SCALE);
  pango_font_description_set_weight(ms_df, PANGO_WEIGHT_NORMAL);
  ms_attr = pango_attr_font_desc_new(ms_df);
  pango_font_description_free(ms_df);
  ms_attr->start_index = 0;
  ms_attr->end_index = -1;
  pango_attr_list_insert(ms_attrlist, ms_attr);

  // set monospacel bold font attributes
  bld_attrlist = pango_attr_list_new();
  bld_df = pango_font_description_new();
  pango_font_description_set_family(bld_df, "Mono");
  pango_font_description_set_size(bld_df, 10*PANGO_SCALE);
  pango_font_description_set_weight(bld_df, PANGO_WEIGHT_BOLD);
  bld_attr = pango_attr_font_desc_new(bld_df);
  pango_font_description_free(bld_df);
  bld_attr->start_index = 0;
  bld_attr->end_index = -1;
  pango_attr_list_insert(bld_attrlist, bld_attr);


  // console messages
  console_messages.messages = gtk_text_buffer_new(0);
  gtk_text_buffer_create_tag(console_messages.messages, "monospace", "family", "Mono", 0);
  gtk_text_buffer_get_start_iter(console_messages.messages, &iter);
  console_messages.mark = gtk_text_buffer_create_mark(console_messages.messages, 0, &iter, 0);

  // new window
  window = gtk_application_window_new(app);
  gtk_window_set_title(GTK_WINDOW(window), "Instant messaging ohjemla");
  gtk_window_set_default_size(GTK_WINDOW(window), 600, 400);

  g_signal_connect(window, "destroy", G_CALLBACK(gtk_main_quit), NULL);

  // vertical container box for tabs and server_container
  main_container = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
  //tabs
  tabs = gtk_notebook_new();
  gtk_box_pack_start((GtkBox*)main_container, tabs, 0, 0, 5);

  // container box for server
  server_container = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
  // channel list
  channel_list_container = gtk_scrolled_window_new(0, 0);
  gtk_scrolled_window_set_policy((GtkScrolledWindow*)channel_list_container, GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
  channel_list = gtk_list_box_new();
  gtk_container_add(GTK_CONTAINER(channel_list_container), channel_list);
  gtk_box_pack_start((GtkBox*)server_container, channel_list_container, 0, 0, 5);
  gtk_widget_set_size_request(channel_list, 150, 0);
  // container box for messages & message input & user list
  channel_container = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 5);

  // container box for topic&messages container and callbuttons&my_message container
  messages_container = gtk_box_new(GTK_ORIENTATION_VERTICAL, 5);

  // container for topic && messages
  topic_messages_container = gtk_box_new(GTK_ORIENTATION_VERTICAL, 5);
  // label for channel topic
  topicline = gtk_label_new("Loading...");
  gtk_label_set_line_wrap((GtkLabel*)topicline, 1);
  // message box
  messages_box = gtk_text_view_new();

  gtk_box_pack_start((GtkBox*)topic_messages_container, topicline, 0, 0, 5);

  // wrap words
  gtk_text_view_set_wrap_mode((GtkTextView*)messages_box, GTK_WRAP_WORD_CHAR);
  // set scrollable
  messages_scrollframe = gtk_scrolled_window_new(0, 0);
  gtk_scrolled_window_set_policy((GtkScrolledWindow*)messages_scrollframe, GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
  gtk_container_add(GTK_CONTAINER(messages_scrollframe), messages_box);
  gtk_text_view_set_editable((GtkTextView*)messages_box, 0);
//  gtk_box_pack_start((GtkBox*)messages_container, messages_scrollframe, 1, 1, 5);

  gtk_box_pack_end((GtkBox*)topic_messages_container, messages_scrollframe, 1, 1, 5);

  gtk_box_pack_start((GtkBox*)messages_container, topic_messages_container, 1, 1, 5);

  // container for call buttons and message input
  callbuttons_mymessage_container = gtk_box_new(GTK_ORIENTATION_VERTICAL, 5);
  callbuttons = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 5);
  // end call button
  endcall_button = gtk_button_new_with_label("End call");
  g_signal_connect(endcall_button, "clicked", G_CALLBACK(hangup), 0);
  gtk_box_pack_start((GtkBox*)callbuttons, endcall_button, 1, 1, 5);

  gtk_box_pack_start((GtkBox*)callbuttons_mymessage_container, callbuttons, 0, 0, 5);

  // message input
  my_message = gtk_entry_new();
  gtk_box_pack_end((GtkBox*)messages_container, callbuttons_mymessage_container, 0, 0, 5);

  gtk_box_pack_end((GtkBox*)callbuttons_mymessage_container, my_message, 1, 1, 5);

  // connect ENTER keypress to parse_user_command()
  g_signal_connect(my_message, "activate", G_CALLBACK(parse_user_command), 0);
  // connect channel selection to select_channel()
  g_signal_connect(channel_list, "row_activated", G_CALLBACK(select_channel), 0);

  // add message container to channel container
  gtk_box_pack_start((GtkBox*)channel_container, messages_container, 1, 1, 5);

  // user list
  nick_list_container = gtk_scrolled_window_new(0, 0);
  gtk_scrolled_window_set_policy((GtkScrolledWindow*)nick_list_container, GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
  nick_list = gtk_list_box_new();
  gtk_container_add(GTK_CONTAINER(nick_list_container), nick_list);
  gtk_box_pack_end((GtkBox*)channel_container, nick_list_container, 0, 0, 5);
  gtk_widget_set_size_request(nick_list, 150, 0);

  // connect nick list selection to select_nick()
  g_signal_connect(nick_list, "row_activated", G_CALLBACK(select_nick), 0);

  // add channel container to server container
  gtk_box_pack_end((GtkBox*)server_container, channel_container, 1, 1, 5);

  // add main tab
  gtk_notebook_append_page((GtkNotebook*)tabs, gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0), gtk_label_new("Command center"));
  // add tabs to window
//  gtk_container_add(GTK_CONTAINER(window), tabs);

  // add server_container to main_container
  gtk_box_pack_end((GtkBox*)main_container, server_container, 1, 1, 5);

  // add container box to window
  gtk_container_add(GTK_CONTAINER(window), main_container);

  // add idle function
//  g_idle_add(idle_function, (gpointer)0);

  gtk_widget_show_all(window);

  // connect server active server selection to select_server()
  g_signal_connect(tabs, "switch-page", G_CALLBACK(select_server), 0);

//  signal(SIGUSR1, idle_function);

  g_application_command_line_get_arguments(command_line, &argc);
  if(argc >= 3)
  {
    active_server = new_connection(g_application_command_line_get_arguments(command_line, 0)[1], atoi(g_application_command_line_get_arguments(command_line, 0)[2]));
    gtk_text_view_set_buffer((GtkTextView*)messages_box, servers[active_server].messages.messages);
  }

  if(argc == 4)
  {
    set_nick(g_application_command_line_get_arguments(command_line, 0)[3]);
    register_user(g_application_command_line_get_arguments(command_line, 0)[3]);
  }
  else
    set_user_message("/nick ");

//  g_idle_add(idle_function, 0);
  eventsrc = g_io_channel_unix_new(pipefd[0]);
  g_io_channel_set_encoding(eventsrc, NULL, &error);
  g_io_channel_set_flags(eventsrc, g_io_channel_get_flags(eventsrc) | G_IO_FLAG_NONBLOCK, &error);
  g_io_add_watch(eventsrc, G_IO_IN | G_IO_PRI, idle_function, 0);

  gtk_widget_show_all(window);

  hide_endcall_button();

  gtk_main();
}

void dumb_signal_handler(int sig)
{
//  char chr;
//  g_idle_add(idle_function, 0);
//  gdk_threads_add_idle(idle_function, 0);
//  g_main_context_invoke(0, idle_function, 0);
//  read(pipefd[0], &chr, 1);
//  idle_function();
//  return 1;
}

int gtk_application(int argc, char **argv)
{
  GtkApplication *app;
  int status;

  app = gtk_application_new("chat.desktop", G_APPLICATION_HANDLES_COMMAND_LINE);
  g_signal_connect(app, "command-line", G_CALLBACK (activate), NULL);

//  signal(SIGUSR1, dumb_signal_handler);

  status = g_application_run(G_APPLICATION(app), argc, argv);

  g_object_unref(app);

  return status;
}

int main(int argc, char **argv)
{
  char *args[] = { "a.out", "sininenankka.dy.fi", "6667", argv[1] };
//  char *args[] = { "a.out", "10.42.0.1", "6667", argv[1] };

  load_settings();

  srand(time(0));
  init_audio();
  main_pid = pthread_self();
  pipe(pipefd);
  if(pthread_mutex_init(&lock, 0))
    abort();

//  return gtk_application(argc, argv);
  return gtk_application(4, args);
}
