void send_raw_command(const char *message, unsigned short len)
{
  char newline[2] = { 0x0D, 0x0A };
  char tcp_payload[MSGLEN];

  if(active_server < 0) return;

  strcpy(tcp_payload, message);
  strncat(tcp_payload, newline, 2);


  send(servers[active_server].socket, tcp_payload, len+2, 0);

}

int get_user_command(const char *cmd, unsigned len)
{
  unsigned n;
  char command[511];

  for(n=len;n--;)
    command[n] = tolower(cmd[n]);

  if(!strncmp(command, "wc", 2))
    return 10;

  if(len >= 3)
  {
    if(!strncmp(command, "msg", 3))
      return 0;
  }
  if(len >= 4)
  {
    if(!strncmp(command, "join", 4))
      return 1;
    if(!strncmp(command, "nick", 4))
      return 2;
    if(!strncmp(command, "pcall", 5))
      return 3;
    if(!strncmp(command, "hangup", 6))
      return 4;
    if(!strncmp(command, "scall", 5))
      return 5;
    if(!strncmp(command, "recdev", 6))
      return 6;
    if(!strncmp(command, "playdev", 6))
      return 7;
    if(!strncmp(command, "part", 4))
      return 8;
    if(!strncmp(command, "query", 5))
      return 9;
  }
  return -1;
}

void send_message(unsigned server, unsigned channel, const char *message)
{
  char command[MSGLEN+1];

  if(channel < servers[server].channel_count)
  {
    sprintf(command, "PRIVMSG %s :%s\r\n", servers[server].channels[channel].name, message);
    send_raw_command(command, strlen(command));
    sprintf(command, "<%s> %s\r\n", servers[server].nick, message);
    push_channel_message(server, command, strlen(command), channel);
  }
  else
  {
    sprintf(command, "PRIVMSG %s :%s\r\n", servers[server].queries[channel-servers[server].channel_count].nick, message);
    send_raw_command(command, strlen(command));
    sprintf(command, "<%s> %s\r\n", servers[server].nick, message);
    push_channel_message(server, command, strlen(command), channel);
  }
}

void parse_user_command(void)
{
  const char *message = gtk_entry_get_text((GtkEntry*)my_message);
  unsigned short len = strlen(message);

  if(message[0] == '/')
    switch(get_user_command(&message[1], len-1))
    {
      case 1:
        join_channel(&message[6], len-6);
      break;
      case 2:
        set_nick(&message[6]);
        if(!servers[active_server].registered)
          register_user(&message[6]);
      break;
      case 3:
        voice_call(&message[6], len-6, 1);
      break;
      case 4:
        hangup();
      break;
      case 5:
        voice_call(&message[6], len-6, 0);
      break;
      case 6:
        set_rec_device(&message[7]);
      break;
      case 7:
        set_play_device(&message[8]);
      break;
      case 8:
        part_channel(&message[6], len-6);
      break;
      case 9:
        add_querytab(active_server, &message[7], 0);
      break;
      case 10:
        window_close(active_server);
      break;
    }
  else if(servers[active_server].active_channel < 0)
    send_raw_command(message, len);
  else
    send_message(active_server, servers[active_server].active_channel, message);

  gtk_entry_set_text((GtkEntry*)my_message, "");
}

int push_channel_message(int server, char *msg, unsigned len, int channel)
{
  void *ptr;
  struct channel_messages *messages;

  if(server < 0)
    messages = &console_messages;
  else if(channel < 0)
    messages = &servers[server].messages;
  else if(channel < servers[server].channel_count)
    messages = &servers[server].channels[channel].messages;
  else
    messages = &servers[server].queries[channel-servers[server].channel_count].messages;

  pthread_mutex_lock(&lock);
  if(!messages->line_queue)
  {
    messages->line_queue = malloc(++messages->line_queue_len * sizeof(struct message));
    if(!messages->line_queue)
    {
      nomemory:
      messages->line_queue_len--;
      pthread_mutex_unlock(&lock);
      return -1;
    }
  }
  else
  {
    ptr = realloc(messages->line_queue, ++messages->line_queue_len * sizeof(struct message));
    if(!ptr) goto nomemory;
    messages->line_queue = ptr;
  }
  messages->line_queue[messages->line_queue_len-1].text = malloc(len * sizeof(char));
  if(!messages->line_queue[messages->line_queue_len-1].text)
    goto nomemory;

  memcpy(messages->line_queue[messages->line_queue_len-1].text, msg, len);
  messages->line_queue[messages->line_queue_len-1].len = len;

  pthread_mutex_unlock(&lock);

  update_ui();

  return 0;
}

int get_channel_message(int server, char *msg, int channel)
{
  unsigned n;
  unsigned q = 0;
  struct channel_messages *messages;

  if(server < 0)
    messages = &console_messages;
  else if(channel < 0)
    messages = &servers[server].messages;
  else if(channel < servers[server].channel_count)
    messages = &servers[server].channels[channel].messages;
  else
    messages = &servers[server].queries[channel-servers[server].channel_count].messages;


  if(!messages->line_queue)
    return -1;

//  pthread_mutex_lock(&lock);

  for(n=0;n<messages->line_queue[0].len;n++)
    msg[n] = messages->line_queue[0].text[n];

  free(messages->line_queue[0].text);

  messages->line_queue_len--;

  while(q < messages->line_queue_len)
  {
    messages->line_queue[q] = messages->line_queue[q+1];
    q++;
  }
  if(!q)
  {
    free(messages->line_queue);
    messages->line_queue = 0;
  }
  else
    messages->line_queue = realloc(messages->line_queue, messages->line_queue_len*sizeof(struct message));

//  pthread_mutex_unlock(&lock);

  return n;
}

void answer_ping(char *command, unsigned len, unsigned server)
{
  command[1] = 'O';
  send(servers[server].socket, command, len, 0);
}

void parse_user_list(unsigned server, char *ptr, unsigned ready)
{
  char channel[CHANLEN];
  unsigned n, m;
  unsigned chanid;
  char user[NICKLEN];
  unsigned len = strnllen(ptr);

  for(n=0;ptr[n] != ' ';n++)
    channel[n] = ptr[n];
  channel[n] = 0;

  if(ptr[++n] == ':') n++;

  chanid = find_channel_by_name(server, channel);
  if(chanid < 0) return;
  printf("Listing users for channel %s\n", channel);

  if(ready)
  {
    servers[server].channels[chanid].user_list_ready = 1;
    return;
  }

  pthread_mutex_lock(&lock);

  if(servers[server].channels[chanid].user_list_ready)
  {
    while(servers[server].channels[chanid].user_count--)
      free(servers[server].channels[chanid].users[servers[server].channels[chanid].user_count]);
    free(servers[server].channels[chanid].users);
  }

  pthread_mutex_unlock(&lock);

  servers[server].channels[chanid].user_list_ready = 0;

  while(n < len)
  {
    m = 0;
    while(ptr[n] > ' ')
      user[m++] = ptr[n++];
    user[m] = 0;
    n++;
    channel_add_nick(server, chanid, user);
  }

  update_ui();
}

int find_server_query(unsigned server, const char *nick)
{
  int n = servers[server].query_count;

  while(n--)
    if(!strcmp(servers[server].queries[n].nick, nick))
      break;

  return n;
}

void receive_query(unsigned server, char *sender, char *message)
{
  int n = find_server_query(server, sender);

  if(n >= 0)
    push_channel_message(server, message, strlen(message), servers[server].channel_count + n);
  else
    add_querytab(server, sender, message);
}

void parse_privmsg(unsigned server, char *command, char *channelptr, unsigned len)
{
  char nick[NICKLEN];
  char channel[CHANLEN];
  unsigned n;
  char *ptr = &command[1];
  char *msgptr;
  char message[MSGLEN];

  for(n=0;channelptr[n] != ' ';n++)
    channel[n] = channelptr[n];
  channel[n] = 0;
  msgptr = channelptr+n+1;

  for(n=0;ptr[n] != '!';n++)
    nick[n] = ptr[n];
  nick[n] = 0;

  sprintf(message, "<%s> ", nick);
  for(n=strlen(message);*msgptr != 0xA;n++)
    message[n] = *(++msgptr);
  message[n] = 0;


  if(!strcmp(channel, servers[server].nick))
    receive_query(server, nick, message);
  else
    push_channel_message(server, message, strlen(message), find_channel_by_name(server, channel));
}

int parse_command(unsigned server, char *command, unsigned len, char **cptr)
{
  char *ptr;

  if(command[0] == ':')
  {
    if(ptr = strnchr(&command[1], ' ', len-1))
    {
      if(!strncmp(ptr, "JOIN", 4))
      {
        *cptr = ptr + 5;
        if(**cptr == ':') (*cptr)++;
        if(!strncmp(&command[1], servers[active_server].nick, strlen(servers[active_server].nick)))
          return 1; // join
        else
          return 5;
      }
      if(!strncmp(ptr, "PART", 4))
      {
       	*cptr = ptr + 5;
        if(**cptr == ':') (*cptr)++;
        if(!strncmp(&command[1], servers[active_server].nick, strlen(servers[active_server].nick)))
          return 6; // part
	else
          return 7;
      }
      else if(!strncmp(ptr, "PRIVMSG", 7))
      {
        *cptr = ptr+8;
        return 2;
      }
      else if(!strncmp(ptr, "TOPIC", 5))
      {
        *cptr = &command[1];
        return 13;
      }
      else if(!strncmp(ptr, "353", 3))
      {
        *cptr = strstr(ptr, servers[active_server].nick) + strlen(servers[active_server].nick) + 3;
        return 3; // names list
      }
      else if(!strncmp(ptr, "366", 3))
      {
        *cptr = strstr(ptr, servers[server].nick) + strlen(servers[server].nick) + 1;
        return 4; // end of names
      }
      else if(!strncmp(ptr, "332", 3))
      {
        *cptr = ptr+4;
        return 12; // topic
      }
    }
  }

  if(len >= 4)
  {
    if(!strncmp(command, "PING", 4))
      return 0;
  }

  if(len >= 7)
  {
    if(!strncmp(command, "HANGUP", 6))
    {
      *cptr = command+7;
      return 10;
    }
  }

  if(len >= 9)
  {
    if(!strncmp(command, "VOIPJOIN", 8))
    {
      *cptr = command+9;
      return 11;
    }
  }

  if(len >= 10)
  {
    if(!strncmp(command, "VOICEPORT", 9))
    {
      *cptr = command+10;
      return 8;
    }
    if(!strncmp(command, "VOIPUDP", 7))
    {
      *cptr = command+8;
      return 9;
    }
  }

  return -1;
}

int command_handler(char *command, unsigned len, unsigned server)
{
  char *ptr;
  int command_type = parse_command(server, command, len, &ptr);

  switch(command_type)
  {
    case 0:
      answer_ping(command, len, server);
    break;
    case 1:
      add_channel(server, ptr);
      break;
    case 2:
      parse_privmsg(server, command, ptr, len);
      break;
    case 3:
      parse_user_list(server, ptr, 0);
      break;
    case 4:
      parse_user_list(server, ptr, 1);
      break;
    case 5:
      add_user_to_channel(server, &command[1], ptr);
      break;
    case 7:
      remove_user_from_channel(server, &command[1], ptr);
      break;
    case 8:
      connect_voice_port(server, ptr);
      break;
    case 9:
      add_voice_recp(ptr);
      break;
    case 10:
      if(servers[server].joining_p2p_voice)
        del_voice_recp(ptr);
      else
        del_svoice_user(ptr);
      update_users();
      break;
    case 11:
      add_svoice_user(ptr);
      break;
    case 12:
      channel_set_topic(server, ptr);
      break;
    case 13:
      channel_reset_topic(server, ptr);
      break;
    case -1: default:
      push_channel_message(server, command, len, -1);
  }
}
