//#define SAMPLES 0x800
#define SAMPLES 2880
#define AUDIO_FORMAT AUDIO_S16SYS

struct audio_buffer
{
  void *volatile buff;
  volatile unsigned len;
  volatile unsigned seq_num;
  volatile unsigned ack_num;
  volatile unsigned played;
  SDL_AudioFormat format;
  unsigned char *frame[4];
};

char **list_audio_devices(int *count, unsigned input)
{
  unsigned n;
  char **retval = 0;

  *count = SDL_GetNumAudioDevices(input);
  if(*count >= 0)
  {
    retval = malloc(*count * sizeof(char*));
    for(n=0;n<*count;n++)
    {
      retval[n] = malloc((strlen(SDL_GetAudioDeviceName(n, input))+1) * sizeof(char));
      strcpy(retval[n], SDL_GetAudioDeviceName(n, input));
    }
  }

  return retval;
}

unsigned audio_device_count(unsigned input)
{
  return SDL_GetNumAudioDevices(input);
}

void free_audio_device_list(char **list, unsigned count)
{
  while(count--)
    free(list[count]);
  free(list);
}

void init_audio(void)
{
  if(SDL_Init(SDL_INIT_AUDIO))
    abort();
}

void audio_input_callback(void *userdata, unsigned char *stream, int len)
{
 ((struct audio_buffer*)userdata)->seq_num++;
  ((struct audio_buffer*)userdata)->played = 0;
  memcpy((void*)((struct audio_buffer*)userdata)->buff, stream, len);
}

unsigned open_audio_input(unsigned device, struct audio_buffer *buffer)
{
  struct SDL_AudioSpec desired_format;
  struct SDL_AudioSpec obtained_format;
  SDL_AudioDeviceID dev;

  desired_format.freq = 48000;
  desired_format.format = AUDIO_FORMAT;
  desired_format.channels = 1;
  desired_format.samples = SAMPLES;
  desired_format.callback = audio_input_callback;
  desired_format.userdata = buffer;

  dev = SDL_OpenAudioDevice(SDL_GetAudioDeviceName(device, 1), 1, &desired_format, &obtained_format, 0);

  if(!dev)
    puts("Error opening audio device");

  printf("Opened device %i\n", dev);
  printf("Freq: %u\n", obtained_format.freq);
  printf("Format: %X\n", obtained_format.format);
  printf("Channels: %u\n", obtained_format.channels);
  printf("Samples: %u\n", obtained_format.samples);
  printf("Silence value: %u\n", obtained_format.silence);
  printf("Buffer size: %u\n", obtained_format.size);

  buffer->buff = malloc(obtained_format.size);
  buffer->len = obtained_format.size;
  buffer->format = obtained_format.format;

  return dev;
}

void audio_output_callback(void *userdata, unsigned char *stream, int len)
{
  struct audio_buffer local_buffer_struct = *(struct audio_buffer*)userdata;
//  unsigned copied;
  unsigned buffer_index = local_buffer_struct.ack_num & 3;

  memcpy(stream, (void*)local_buffer_struct.frame[buffer_index], len);
  memset((void*)local_buffer_struct.frame[buffer_index], 0, len);

  if(((struct audio_buffer*)userdata)->ack_num < ((struct audio_buffer*)userdata)->seq_num)
    ((struct audio_buffer*)userdata)->ack_num++;
  else
    ((struct audio_buffer*)userdata)->ack_num-=4;

}

unsigned open_audio_output(unsigned device, struct audio_buffer *buff)
{
  struct SDL_AudioSpec desired_format;
  struct SDL_AudioSpec obtained_format;
  SDL_AudioDeviceID dev;

  desired_format.freq = 48000;
  desired_format.format = AUDIO_FORMAT;
  desired_format.channels = 1;
  desired_format.samples = SAMPLES;
  desired_format.callback = audio_output_callback;
  desired_format.userdata = buff;

  dev = SDL_OpenAudioDevice(SDL_GetAudioDeviceName(device, 0), 0, &desired_format, &obtained_format, 0);

  if(!dev)
    puts("Error opening audio device");

  printf("Opened device %i\n", dev);
  printf("Freq: %u\n", obtained_format.freq);
  printf("Format: %X\n", obtained_format.format);
  printf("Channels: %u\n", obtained_format.channels);
  printf("Samples: %u\n", obtained_format.samples);
  printf("Silence value: %u\n", obtained_format.silence);
  printf("Buffer size: %u\n", obtained_format.size);

  buff->frame[0] = buff->buff = calloc(obtained_format.size*4, 1);
  buff->len = obtained_format.size;
  buff->format = obtained_format.format;
  buff->frame[1] = buff->frame[0] + obtained_format.size;
  buff->frame[2] = buff->frame[1] + obtained_format.size;
  buff->frame[3] = buff->frame[2] + obtained_format.size;

  return dev;
}

void play_audio_device(unsigned dev)
{
  SDL_PauseAudioDevice(dev, SDL_FALSE);
}

void close_audio_device(unsigned dev)
{
  SDL_LockAudioDevice(dev);
  SDL_CloseAudioDevice(dev);
}

void pause_audio_device(unsigned dev)
{
  SDL_PauseAudioDevice(dev, SDL_TRUE);
}
