GtkWidget *settings_window;
GtkWidget *settings_main_container;
GtkWidget *silence_value_container;
GtkWidget *settings_close_button;
GtkWidget *silence_scrollbar;
GtkWidget *talking_indicator;
GtkAdjustment *silence_value;
struct audio_buffer settings_mic_buff;
int settings_mic_dev;

void settings_update_min_volume(void)
{
  MIN_VOLUME = gtk_adjustment_get_value((GtkAdjustment*)silence_value);
  MIN_VOLUME = (MIN_VOLUME) / (0x100-(MIN_VOLUME >> 8));
}

int settings_idle(void *arg)
{
  if(average_volume((unsigned short*)settings_mic_buff.buff, SAMPLES) >= MIN_VOLUME)
    gtk_label_set_markup((GtkLabel*)talking_indicator, "<span foreground='green'>Talking</span>");
  else
    gtk_label_set_markup((GtkLabel*)talking_indicator, "<span foreground='black'>Talking</span>");
}

void settings_listen_mic(void)
{
  settings_mic_dev = open_audio_input(recording_device, &settings_mic_buff);
  play_audio_device(settings_mic_dev);
}

void settings_window_closed(void)
{
  close_audio_device(settings_mic_dev);
  g_idle_remove_by_data(0);
  free(settings_mic_buff.buff);
}

void close_settings_window_and_save(void)
{
  gtk_widget_destroy(settings_window);
  save_settings();
}

void open_settings_window(void)
{
  settings_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title((GtkWindow*)settings_window, "Settings");
  gtk_window_set_default_size((GtkWindow*)settings_window, 400, 200);

  g_signal_connect(settings_window, "destroy", G_CALLBACK(settings_window_closed), 0);

  settings_main_container = gtk_box_new(GTK_ORIENTATION_VERTICAL, 5);

  silence_value_container = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 5);

  settings_close_button = gtk_button_new_with_label("Close and save");

  g_signal_connect(settings_close_button, "clicked", G_CALLBACK(close_settings_window_and_save), 0);

  silence_value = gtk_adjustment_new(MIN_VOLUME * (0x100-(MIN_VOLUME >> 8)), 0, 0xFFFF, 0, 0, 0);

  silence_scrollbar = gtk_scrollbar_new(GTK_ORIENTATION_HORIZONTAL, silence_value);

  g_signal_connect(silence_scrollbar, "value-changed", G_CALLBACK(settings_update_min_volume), 0);

  talking_indicator = gtk_label_new("Talking");

  gtk_box_pack_start((GtkBox*)silence_value_container, talking_indicator, 0, 0, 0);
  gtk_box_pack_end((GtkBox*)silence_value_container, silence_scrollbar, 1, 1, 0);

  gtk_box_pack_start((GtkBox*)settings_main_container, silence_value_container, 0, 0, 0);
  gtk_box_pack_end((GtkBox*)settings_main_container, settings_close_button, 0, 0, 0);

  gtk_container_add((GtkContainer*)settings_window, settings_main_container);

  settings_listen_mic();

  gtk_widget_show_all(settings_window);

  g_idle_add(settings_idle, 0);
}
