#define MIN_VOLUME min_input_volume
#define OPUS_SAMPLES SAMPLES
#define BUSYLOOP 2

void cancel_listen(void *arg)
{
  unsigned server_id = (unsigned)(long)arg;

  puts("Canceling VoIP listen thread...");
  close_audio_device(servers[server_id].play_dev);
  free(servers[server_id].input_buff.buff);
  puts("Closed playback device");
  free(servers[server_id].mix_buffer);
  puts("Freed mix buffer");
  free(servers[server_id].input_buffer);
  puts("Freed input buffer");
  opus_decoder_destroy(servers[server_id].opus_decoder);
  puts("Destroyed opus decoder");

  puts("VoIP listen thread canceled");
}

void setup_encoder(OpusEncoder *opus_encoder)
{
//  opus_encoder_ctl(opus_encoder, OPUS_SET_BITRATE(OPUS_BITRATE_MAX));
//  opus_encoder_ctl(opus_encoder, OPUS_SET_DTX(1));
}

unsigned ms_to_time(long int nanoseconds_now, long int *nanoseconds_to)
{
  while(*nanoseconds_to > 999999999)
    *nanoseconds_to -= 1000000000;
  if(*nanoseconds_to > nanoseconds_now)
    return (*nanoseconds_to - nanoseconds_now) / 1000000;
  else
    return ((*nanoseconds_to + 1000000000) - nanoseconds_now) / 1000000;
}

char *speaking_p2p_nick(struct sockaddr_in *p2p_addr)
{
  unsigned n = voip_target_count;
  static char retstr[NICKLEN+1];

  while(n--)
  {
    if(p2p_addr->sin_addr.s_addr == voip_targets[n].addr_struct.sin_addr.s_addr
       && p2p_addr->sin_port == voip_targets[n].addr_struct.sin_port)
    {
      strcpy(retstr, voip_targets[n].nick);
      return retstr;
    }
  }
  return 0;
}

void *listen_voip(void *arg)
{
  unsigned server_id = (unsigned)(long)arg;
  int socket = servers[server_id].voice_socket;
  struct sockaddr_in *socket_struct = &servers[server_id].voice_socket_local;
  socklen_t sockaddr_len;
  int n, m;
  struct audio_buffer *input_buff = &servers[server_id].input_buff;
  unsigned outdev = servers[server_id].play_dev;
  unsigned char *mix_buffer = 0;
  unsigned char *input_buffer = 0;
  SDL_AudioFormat mix_format;
  char speaking_nick[NICKLEN];
  unsigned char *ptr;
  OpusDecoder *opus_decoder;
  int opus_error;
  unsigned char *raw_pcm;
  int opus_samples;
  unsigned char *speaker_ptr[4];
  unsigned mix_buffer_index = 0;
  unsigned speaker_buffer_index = 0;
  unsigned packets_received;

  struct pollfd read_poll[1];

  long time_before_next_sample;
  long sample_time = (1000000000 / 48000 * SAMPLES); // in nanoseconds
  long next_sample;
  struct timespec nanotime;


  read_poll[0].fd = socket;
  read_poll[0].events = POLLIN;

  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, 0);

  if(SDL_SetThreadPriority(SDL_THREAD_PRIORITY_HIGH))
    printf("Error when trying to set VoIP thread priority: %s\n", SDL_GetError());

  puts("Creating Opus decoder...");
  servers[server_id].opus_decoder = opus_decoder = opus_decoder_create(48000, 1, &opus_error);
  if(opus_error) printf("Opus error: %s\n", opus_strerror(opus_error));

  pthread_cleanup_push(&cancel_listen, arg);

  play_audio_device(outdev);
  mix_format = input_buff->format;

  servers[server_id].mix_buffer = mix_buffer = malloc(input_buff->len*sizeof(char));
  servers[server_id].raw_pcm = raw_pcm = malloc(input_buff->len * sizeof(char));
  if(servers[server_id].joining_p2p_voice)
    servers[server_id].input_buffer = input_buffer = malloc(input_buff->len * sizeof(char));
  else
    servers[server_id].input_buffer = input_buffer = malloc((input_buff->len+NICKLEN+1) * sizeof(char));

  if(!mix_buffer || !input_buffer || !raw_pcm)
  {
    free(mix_buffer);
    free(input_buffer);
    puts("Out of memory!");
    return 0;
  }

  puts("Listening UDP...");

  socket_struct->sin_addr.s_addr = INADDR_ANY;

  clock_gettime(CLOCK_REALTIME, &nanotime);
  next_sample = nanotime.tv_nsec;

  while(1)
  {
//    printf("Trying to receive %u long packet to port %u\n", input_buff.len, ntohs(servers[server_id].voice_socket_local.sin_port));
    input_buff->played = 0;
    memset(mix_buffer, 0, input_buff->len);
    do
    {
      next_sample += sample_time;

      clock_gettime(CLOCK_REALTIME, &nanotime);
      time_before_next_sample = ms_to_time(nanotime.tv_nsec, &next_sample);
    }
    while(time_before_next_sample*1000000 > sample_time); // if overflowed because did not get enough CPU time

    if(time_before_next_sample < BUSYLOOP) time_before_next_sample=BUSYLOOP;

    if(servers[server_id].joining_p2p_voice)
    {
      while(input_buff->seq_num > input_buff->ack_num+3)
      {
        if(!poll(read_poll, 1, time_before_next_sample-BUSYLOOP))
          while(recvfrom(socket, input_buffer, input_buff->len, MSG_PEEK | MSG_DONTWAIT, (struct sockaddr*)socket_struct, &sockaddr_len) < 0)
          {
            clock_gettime(CLOCK_REALTIME, &nanotime);
            time_before_next_sample = ms_to_time(nanotime.tv_nsec, &next_sample);
            if(!time_before_next_sample)
              break;
          }

        sockaddr_len = sizeof(struct sockaddr_in);
        n = recvfrom(socket, input_buffer, input_buff->len, 0, (struct sockaddr*)socket_struct, &sockaddr_len);
        if(n <= 0)
        {
          if(errno != EWOULDBLOCK)
            printf("Listen: %s\n", strerror(errno));
          else
          {
            continue;
          }
        }
        else if(n == 1 && input_buffer[0] == 0) // keepalive packet
          continue;
        else
        {
          if(ptr = speaking_p2p_nick(socket_struct))
          {
            if((opus_error = opus_decode(opus_decoder, input_buffer, n, (unsigned short*)raw_pcm, OPUS_SAMPLES, 0)) < 0)
              printf("Opus decoding error %s\n", opus_strerror(opus_error));
            if(!speaking_name)
            {
              speaking_name = ptr;
//              puts(speaking_name);
              update_users();
            }
          }
        }
        SDL_MixAudioFormat(mix_buffer, raw_pcm, mix_format, input_buff->len, SDL_MIX_MAXVOLUME>>1);
      }
      memcpy((unsigned char*)input_buff->frame[input_buff->seq_num&3], mix_buffer, input_buff->len);
    }
    else
    {
      while(input_buff->seq_num > input_buff->ack_num+3)
      {
        if(!poll(read_poll, 1, time_before_next_sample-BUSYLOOP))
          while(recvfrom(socket, input_buffer, input_buff->len+NICKLEN+1, MSG_PEEK | MSG_DONTWAIT, (struct sockaddr*)socket_struct, &sockaddr_len) < 0)
          {
            clock_gettime(CLOCK_REALTIME, &nanotime);
            time_before_next_sample = ms_to_time(nanotime.tv_nsec, &next_sample);
            if(!time_before_next_sample)
              break;
          }

        sockaddr_len = sizeof(struct sockaddr_in);
        n = recvfrom(socket, input_buffer, input_buff->len+NICKLEN+1, 0, (struct sockaddr*)socket_struct, &sockaddr_len);
        if(n <= 0)
        {
          if(errno != EWOULDBLOCK)
            printf("Listen: %s\n", strerror(errno));
          else
          {
            continue;
          }
        }
//        printf("Voice packet, buffer usage: %u\n", input_buff->seq_num - input_buff->ack_num);
        ptr = input_buffer;
        while(*ptr++);
        memcpy(speaking_nick, input_buffer, ptr-input_buffer);
        if(!speaking_name)
        {
          speaking_name = speaking_nick;
          update_users();
        }
        if((opus_error = opus_decode(opus_decoder, ptr, n, (unsigned short*)raw_pcm, OPUS_SAMPLES, 0)) < 0)
            printf("Opus decoding error %s\n", opus_strerror(opus_error));
        SDL_MixAudioFormat(mix_buffer, raw_pcm, mix_format, input_buff->len, SDL_MIX_MAXVOLUME>>1);
      }
      memcpy((unsigned char*)input_buff->frame[input_buff->seq_num&3], mix_buffer, input_buff->len);
    }
    ++input_buff->seq_num;
  }

  pthread_cleanup_pop(0);
}

void cancel_send(void *arg)
{
  unsigned server_id = (unsigned)(long)arg;

  puts("Canceling VoIP send thread...");

//  pause_audio_device(servers[server_id].rec_dev);
  close_audio_device(servers[server_id].rec_dev);
  free(servers[server_id].output_buff.buff);
  puts("Closed recording device");

  opus_encoder_destroy(servers[server_id].opus_encoder);
  puts("Destroyed opus encoder");

  puts("VoIP send thread canceled");
}

unsigned short average_volume(short *raw_audio, unsigned samples)
{
  short min = 0xFFFF;
  short max = 0;

//  while(samples-=0x10)
  while(samples--)
  {
    if(min > raw_audio[samples]) min = raw_audio[samples];
    if(max < raw_audio[samples]) max = raw_audio[samples];
  }

  return max-min;
}

unsigned send_frames = 0;

void *send_voip(void *arg)
{
  unsigned n;
  unsigned server_id = (unsigned)(long)arg;
//  int s = servers[server_id].voice_send_socket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  int s = servers[server_id].voice_socket;
  socklen_t sockaddr_len = sizeof(struct sockaddr_in);
  struct audio_buffer *output_buff = &servers[server_id].output_buff;
  unsigned indev = servers[server_id].rec_dev;
  unsigned seq_num = 0;
  OpusEncoder *opus_encoder;
  int opus_error;
  unsigned char *opus_data;
  int opus_length;
  unsigned long last_packet;
  unsigned char *raw_packet;

  long time_before_next_sample;
  long sample_time = (1000000000 / 48000 * SAMPLES); // in nanoseconds
  long next_sample;
  struct timespec nanotime;

  if(s < 0)
  {
    puts("Error while opening socket for sending VoIP datagrams");
    return 0;
  }

  if(!bind(s, (struct sockaddr*)&servers[server_id].voice_socket_local, sockaddr_len))
  {
    puts("errrrorr");
    return 0;
  }

  puts("Creating Opus encoder...");
  servers[server_id].opus_encoder = opus_encoder = opus_encoder_create(48000, 1, OPUS_APPLICATION_VOIP, &opus_error);
  if(opus_error)
    printf("Opus error: %s\n", opus_strerror(opus_error));

  setup_encoder(opus_encoder);

  pthread_cleanup_push(&cancel_send, arg);

  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, 0);

  if(SDL_SetThreadPriority(SDL_THREAD_PRIORITY_HIGH))
    printf("Error when trying to set VoIP thread priority: %s\n", SDL_GetError());


  if(!(opus_data = malloc(output_buff->len)))
  {
    puts("Out of memory");
    return 0;
  }

  play_audio_device(indev);

  clock_gettime(CLOCK_REALTIME, &nanotime);
  next_sample = nanotime.tv_nsec;

  while(1)
  {
    next:while(seq_num == output_buff->seq_num) do
    {
      next_sample += sample_time;
      clock_gettime(CLOCK_REALTIME, &nanotime);
      time_before_next_sample = ms_to_time(nanotime.tv_nsec, &next_sample);
//printf("%i\n", time_before_next_sample);
      if(time_before_next_sample > BUSYLOOP && time_before_next_sample*1000000 < sample_time)
        usleep((time_before_next_sample-BUSYLOOP)*1000);
    }
    while(time_before_next_sample*1000000 > sample_time); // if overflowed because did not get enough CPU time

    if(time_before_next_sample > BUSYLOOP)
      usleep((time_before_next_sample-BUSYLOOP)*1000);

    while(n = voip_target_count)
    {
//      if(seq_num == output_buff->seq_num)
//        continue;
      if(average_volume((unsigned short*)output_buff->buff, SAMPLES) < MIN_VOLUME)
      {
        if(send_frames) goto sendframe;
        else if(time(0) < last_packet + VOIP_KEEPALIVE)
          goto next;
        else
        {
          memset(output_buff->buff, 0, output_buff->len);
          raw_packet = output_buff->buff;
          opus_length = 1;
          update_users();
        }
      }
      else
      {
        send_frames = 0x10;
        sendframe:send_frames--;
        raw_packet = opus_data;
        seq_num = output_buff->seq_num;
        opus_length = opus_encode(opus_encoder, (unsigned short*)output_buff->buff, OPUS_SAMPLES, opus_data, output_buff->len);
        if(opus_length<0)
          printf("Opus encoding error: %s\n", opus_strerror(opus_length));
      }
      while(n--)
      {
        if(sendto(s, raw_packet, opus_length, 0, (struct sockaddr*)&voip_targets[n].addr_struct, sockaddr_len) <= 0)
          printf("Send: %s\n", strerror(errno));
        else
          last_packet = time(0);
      }
      break;
    }
  }
  pthread_cleanup_pop(0);
}

void *send_svoip(void *arg)
{
  unsigned server_id = (unsigned)(long)arg;
  int s = servers[server_id].voice_socket;
  socklen_t sockaddr_len = sizeof(struct sockaddr_in);
  struct audio_buffer *output_buff = &servers[server_id].output_buff;
  unsigned indev = servers[server_id].rec_dev;
  unsigned seq_num = 0;
  OpusEncoder *opus_encoder;
  int opus_error;
  unsigned char *opus_data;
  int opus_length;
  unsigned long last_packet;
  unsigned char *raw_packet;

  long time_before_next_sample;
  long sample_time = (1000000000 / 48000 * SAMPLES); // in nanoseconds
  long next_sample;
  struct timespec nanotime;

  pthread_cleanup_push(&cancel_send, arg);

  puts("Creating Opus encoder...");
  servers[server_id].opus_encoder = opus_encoder = opus_encoder_create(48000, 1, OPUS_APPLICATION_VOIP, &opus_error);
  if(opus_error)
    printf("Opus error: %s\n", opus_strerror(opus_error));

  setup_encoder(opus_encoder);

  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, 0);

  if(SDL_SetThreadPriority(SDL_THREAD_PRIORITY_HIGH))
    printf("Error when trying to set VoIP thread priority: %s\n", SDL_GetError());

  if(!(opus_data = malloc(output_buff->len)))
  {
    puts("Out of memory");
    return 0;
  }

  play_audio_device(indev);

  last_packet = time(0);

  clock_gettime(CLOCK_REALTIME, &nanotime);
  next_sample = nanotime.tv_nsec;

  while(1)
  {
    while(seq_num == output_buff->seq_num) do
    {
      next_sample += sample_time;
      clock_gettime(CLOCK_REALTIME, &nanotime);
      time_before_next_sample = ms_to_time(nanotime.tv_nsec, &next_sample);
//printf("%i\n", time_before_next_sample);
      if(time_before_next_sample > BUSYLOOP && time_before_next_sample*1000000 < sample_time)
        usleep((time_before_next_sample-BUSYLOOP)*1000);
    }
    while(time_before_next_sample*1000000 > sample_time); // if overflowed because did not get enough CPU time

    if(time_before_next_sample > BUSYLOOP)
      usleep((time_before_next_sample-BUSYLOOP)*1000);

//    while(seq_num == output_buff->seq_num);

    if(average_volume((unsigned short*)output_buff->buff, SAMPLES) < MIN_VOLUME)
    {
      if(send_frames) goto sendframe;
      else if(time(0) < last_packet + VOIP_KEEPALIVE)
        continue;
      else
      {
        memset(output_buff->buff, 0, output_buff->len);
        raw_packet = output_buff->buff;
        opus_length = 1;
        update_users();
      }
    }
    else
    {
      send_frames = 0x10;
      sendframe:send_frames--;
      seq_num = output_buff->seq_num;
      raw_packet = opus_data;
      opus_length = opus_encode(opus_encoder, (unsigned short*)output_buff->buff, OPUS_SAMPLES, opus_data, output_buff->len);
      if(opus_length<0)
        printf("Opus encoding error: %s\n", opus_strerror(opus_length));
    }
    if(sendto(s, raw_packet, opus_length, 0, (struct sockaddr*)&servers[server_id].voice_socket_remote, sockaddr_len) <= 0)
      printf("Send: %s\n", strerror(errno));
    else
      last_packet = time(0);

  }

  pthread_cleanup_pop(0);
}

void connect_voice_port(unsigned server, char *str)
{
  int s;
  socklen_t sockaddr_len = sizeof(struct sockaddr_in);
  char command[MSGLEN];
  unsigned port = strtoul(str, &str, 10);

  if(!servers[server].joining_p2p_voice && !servers[server].joining_server_voice)
    return;

  if((s = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
  {
    puts("Error opening UDP socket");
    return;
  }

  show_endcall_button();

  str++;

  do
  {
    sockaddr_len = sizeof(struct sockaddr_in);
    getsockname(servers[server].socket, (struct sockaddr*)&servers[server].voice_socket_local, &sockaddr_len);
    servers[server].voice_socket_local.sin_port = rand();
  }
  while(bind(s, (struct sockaddr*)&servers[server].voice_socket_local, sockaddr_len));

  getpeername(servers[server].socket, (struct sockaddr*)&servers[server].voice_socket_remote, &sockaddr_len);
  servers[server].voice_socket_remote.sin_port = htons(port);
  servers[server].voice_socket = s;

  str[strnllen(str)] = 0;

  sprintf(command, "%s", str);

  sendto(s, command, strlen(command), 0, (struct sockaddr*)&servers[server].voice_socket_remote, sockaddr_len);

  puts("Opening audio input device...");
  servers[server].rec_dev = open_audio_input(recording_device, &servers[server].output_buff);

  puts("Opening audio output device...");
  servers[server].play_dev = open_audio_output(playback_device, &servers[server].input_buff);


  // create new thread for listening udp
  if(pthread_create(&servers[server].voice_recv_thread, 0, &listen_voip, (void*)(long)server))
    puts("Could not create thread for listening VoIP");

  // create new thread for sending udp
  if(servers[server].joining_p2p_voice)
  {
    if(pthread_create(&servers[server].voice_send_thread, 0, &send_voip, (void*)(long)server))
      puts("Could not create thread for sending VoIP");
  }
  else
    if(pthread_create(&servers[server].voice_send_thread, 0, &send_svoip, (void*)(long)server))
      puts("Could not create thread for sending VoIP");

  puts("Voice connected");
}

void add_voice_recp(char *nickaddrport)
{
  void *ptr;
  unsigned i = voip_target_count;
  unsigned n = 0;

  if(!voip_target_count)
  {
    voip_targets = malloc(++voip_target_count * sizeof(struct voip_udp_target));
    if(!voip_targets)
    {
      nomemory:
      voip_target_count--;
      puts("Out of memory");
      return;
    }
  }
  else
  {
    ptr = realloc(voip_targets, ++voip_target_count * sizeof(struct voip_udp_target));
    if(!ptr) goto nomemory;
    voip_targets = ptr;
  }

  while(*nickaddrport != ' ')
    voip_targets[i].nick[n++] = *nickaddrport++;
  voip_targets[i].nick[n] = 0;

  voip_targets[i].addr_struct.sin_family = AF_INET;
  voip_targets[i].addr_struct.sin_addr.s_addr = inet_addr(++nickaddrport);
  while(*++nickaddrport != ' ');
  voip_targets[i].addr_struct.sin_port = htons(strtoul(++nickaddrport, &nickaddrport, 10));

  printf("Added %s, address %s, port %u to VoIP recipients\n", voip_targets[i].nick, inet_ntoa(voip_targets[i].addr_struct.sin_addr), ntohs(voip_targets[i].addr_struct.sin_port));

}

void del_voice_recp(char *nick)
{
  int n = voip_target_count;

  if(!n) return;

  nick[strnllen(nick)] = 0;

  while(n--)
    if(!strcmp(nick, voip_targets[n].nick))
      break;

  if(n < 0) return;

  voip_target_count--;

  while(n < voip_target_count)
  {
    voip_targets[n] = voip_targets[n+1];
    n++;
  }
  if(voip_target_count)
    voip_targets = realloc(voip_targets, voip_target_count*sizeof(struct voip_udp_target));
  else
  {
    free(voip_targets);
    voip_targets = 0;
  }

  printf("Removed user %s from VoIP targets\n", nick);
}

void add_svoice_user(char *str)
{

}

void del_svoice_user(char *str)
{

}
